# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: too-many-arguments

from collections import defaultdict
from enum import Enum, auto
from typing import List, Optional, Union, cast

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.svc.swan_creator.diagram_creator import DiagramCreator
import ansys.scadeone.core.swan.common as common
import ansys.scadeone.core.swan.scopes as scopes

from .equations import ActivateIf, ActivateWhen, DefByCase, EquationLHS, StateMachine
from .expressions import GroupAdaptation, PortExpr
from .instances import OperatorBase, OperatorExpression


class DiagramObject(common.SwanItem, common.PragmaBase):  # numpydoc ignore=PR01
    """Base class for diagram objects.

    *object* ::= ( [[ *lunum* ]] [[ *luid* ]] *description* [[ *local_objects* ]] )

    Parameters
    ----------
    lunum: Lunum (optional)
        Object local unique number within the current operator.

    luid: Luid (optional)
        Object local unique identifier within the current operator.

    locals: list DiagramObject
        List of local objects associated with the object.
        If locals is None, an empty list is created.
    """

    def __init__(
        self,
        lunum: Optional[common.Lunum] = None,
        luid: Optional[common.Luid] = None,
        locals: Optional[List["DiagramObject"]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        common.SwanItem.__init__(self)
        common.PragmaBase.__init__(self, pragmas)
        self._lunum = lunum
        self._luid = luid
        self._locals = locals if locals else []
        common.SwanItem.set_owner(self, self._locals)

    @property
    def lunum(self) -> Union[common.Lunum, None]:
        """Lunum of object, or None if no Lunum."""
        return self._lunum

    @property
    def luid(self) -> Union[common.Luid, None]:
        """Luid of object, or None if no Luid."""
        return self._luid

    @property
    def locals(self) -> List["DiagramObject"]:
        """Local objects of object."""
        return self._locals

    @property
    def sources(
        self,
    ) -> List[tuple["DiagramObject", GroupAdaptation, Union[List[GroupAdaptation], None]]]:
        """Return a list of all diagram objects that are sources of current diagram object.
        A list item is a tuple of source object and the source and target adaptations used
        for connection if any.
        """
        diagram = self._get_diagram(self.owner)
        return diagram.get_block_sources(self)

    @property
    def targets(
        self,
    ) -> List[tuple["DiagramObject", GroupAdaptation, GroupAdaptation]]:
        """Return a list of all diagram objects that are targets of current diagram object.
        A list item is a tuple of target object and the source and target adaptations used
        for connection if any.
        """
        diagram = self._get_diagram(self.owner)
        return diagram.get_block_targets(self)

    def _get_diagram(self, diag_obj: "DiagramObject") -> "Diagram":
        """Get the diagram from a diagram object."""
        if isinstance(diag_obj, Diagram):
            return diag_obj
        return self._get_diagram(diag_obj.owner)


class Diagram(scopes.ScopeSection, DiagramCreator):  # numpydoc ignore=PR01
    """Class for a **diagram** construct."""

    def __init__(self, objects: List[DiagramObject] = None) -> None:
        super().__init__()
        if objects is None:
            self._objects = []
        else:
            self._objects = objects
        self._diag_nav = None
        common.SwanItem.set_owner(self, objects)

    @property
    def objects(self) -> List[DiagramObject]:
        """Diagram objects."""
        return self._objects

    def get_block_sources(
        self, obj: DiagramObject
    ) -> List[tuple[DiagramObject, Optional[GroupAdaptation], Optional[GroupAdaptation]]]:
        """Return a list of all diagram objects that are sources of current diagram.
        A list item is a tuple of source object and the source and target adaptations used
        for connection if any.
        """
        if self._diag_nav is None:
            self._consolidate()
        return self._diag_nav.get_block_sources(obj)

    def get_block_targets(
        self, obj: DiagramObject
    ) -> List[tuple[DiagramObject, Optional[GroupAdaptation], Optional[GroupAdaptation]]]:
        """Return a list of all diagram objects that are targets of current diagram.
        A list item is a tuple of source object and the source and target adaptations used
        for connection if any.
        """
        if self._diag_nav is None:
            self._consolidate()
        return self._diag_nav.get_block_targets(obj)

    def _consolidate(self) -> None:
        # Retrieves wire sources, wire targets and blocks from the Diagram Object. Internal method.
        self._diag_nav = DiagramNavigation(self)
        self._diag_nav.consolidate()


# Diagram object descriptions
# ------------------------------------------------------------


class ExprBlock(DiagramObject):  # numpydoc ignore=PR01
    """Expression block:

    - *object* ::= ( [[ *lunum* ]] [[ *luid* ]] *description* [[ *local_objects* ]] )
    - *description* ::= **expr** *expr*
    """

    def __init__(
        self,
        expr: common.Expression,
        lunum: Optional[common.Lunum] = None,
        luid: Optional[common.Luid] = None,
        locals: Optional[List["DiagramObject"]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(lunum, luid, locals, pragmas)
        self._expr = expr

    @property
    def expr(self) -> common.Expression:
        """Block expression."""
        return self._expr


class DefBlock(DiagramObject):  # numpydoc ignore=PR01
    """Definition block:

    - *object* ::= ( [[ *lunum* ]]  [[ *luid* ]] *description* [[ *local_objects* ]] )
    - *description* ::= **def** *lhs*
    - *description* ::= **def** {syntax% text %syntax}

    The *is_protected* property returns True when the definition is
    protected with a markup.
    """

    def __init__(
        self,
        lhs: Union[EquationLHS, common.ProtectedItem],
        lunum: Optional[common.Lunum] = None,
        luid: Optional[common.Luid] = None,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(lunum, luid, locals, pragmas)
        self._lhs = lhs
        self._is_protected = isinstance(lhs, str)

    @property
    def lhs(self) -> Union[EquationLHS, common.ProtectedItem]:
        """Returned defined flows."""
        return self._lhs

    @property
    def is_protected(self) -> bool:
        """True when definition is syntactically incorrect and protected."""
        return self._is_protected


class Block(DiagramObject):  # numpydoc ignore=PR01
    """Generic block:

    - *object* ::= ( [[ *lunum* ]] [[ *luid* ]] *description* [[ *local_objects* ]] )
    - *description* ::= **block**  (*operator* | *op_expr* )
    - *description* ::= **block** {syntax% text %syntax}

    The *is_protected* property returns True when the block definition
    is protected with a markup.
    """

    def __init__(
        self,
        instance: Union[OperatorBase, OperatorExpression, common.ProtectedItem],
        lunum: Optional[common.Lunum] = None,
        luid: Optional[common.Luid] = None,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(lunum, luid, locals, pragmas)
        self._instance = instance

    @property
    def instance(self) -> Union[OperatorBase, OperatorExpression, common.ProtectedItem]:
        """Called instance as an Operator, or an OperatorExpression or a protected string."""
        return self._instance

    @property
    def is_protected(self) -> bool:
        """True when called operator is defined as a string."""
        return isinstance(self.instance, common.ProtectedItem)


class Connection(common.SwanItem):  # numpydoc ignore=PR01
    """Wire connection for a source or for targets:

    - *connection* ::= *port* [[ *group_adaptation* ]] | ()

    If both *port* and *adaptation* are None, then it corresponds to the '()' form.

    Connection is not valid if only *adaptation* is given. This is checked
    with the *_is_valid()_* method.
    """

    def __init__(
        self,
        port: Optional[PortExpr] = None,
        adaptation: Optional[GroupAdaptation] = None,
    ) -> None:
        super().__init__()
        self._port = port
        self._adaptation = adaptation

    @property
    def port(self) -> Union[PortExpr, None]:
        """Return the port of the connection."""
        return self._port

    @property
    def adaptation(self) -> Union[GroupAdaptation, None]:
        """Return the adaptation of the port of the connection."""
        return self._adaptation

    @property
    def is_valid(self) -> bool:
        """True when the connection either () or *port* [*adaptation*]."""
        return (self.port is not None) or (self.adaptation is None)

    @property
    def is_connected(self) -> bool:
        """True when connected to some port."""
        return self.is_valid and (self.port is not None)


class Wire(DiagramObject):  # numpydoc ignore=PR01
    """Wire definition:

    - *object* ::= ( [[ *lunum* ]] *description* [[ *local_objects* ]] )
    - *description* ::= **wire** *connection* => *connection* {{ , *connection* }}

    A **wire** *must* have a least one target.
    """

    def __init__(
        self,
        source: Connection,
        targets: List[Connection],
        lunum: Optional[common.Lunum] = None,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(lunum, None, locals, pragmas)
        self._source = source
        self._targets = targets

    @property
    def source(self) -> Connection:
        """Wire source."""
        return self._source

    @property
    def targets(self) -> List[Connection]:
        """Wire targets."""
        return self._targets

    @property
    def has_target(self) -> bool:
        """Return True when wire as at least one target."""
        return len(self.targets) > 0

    @property
    def sources(self) -> List[tuple["DiagramObject", Optional[GroupAdaptation]]]:
        """This method must not be called for a Wire"""
        raise ScadeOneException("Wire.sources() call")


class GroupOperation(Enum):  # numpydoc ignore=PR01
    """Operation on groups."""

    # pylint: disable=invalid-name

    #: No operation on group
    NoOp = auto()

    #: **byname** operation (keep named items)
    ByName = auto()

    #: **bypos** operation (keep positional items)
    ByPos = auto()

    #: Normalization operation (positional, then named items)
    Normalize = auto()

    @staticmethod
    def to_str(value: "GroupOperation"):
        """Group Enum to string."""
        if value == GroupOperation.NoOp:
            return ""
        if value == GroupOperation.Normalize:
            return "()"
        return value.name.lower()


class Bar(DiagramObject):  # numpydoc ignore=PR01
    """Bar (group/ungroup constructor block):

    - *object* ::= ( [[ *lunum* ]] *description* [[ *local_objects* ]] )
    - *description* ::= **group** [[*group_operation*]]
    - *group_operation* ::= () | **byname** | **bypos**
    """

    def __init__(
        self,
        operation: Optional[GroupOperation] = GroupOperation.NoOp,
        lunum: Optional[common.Lunum] = None,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(lunum, None, locals, pragmas)
        self._operation = operation

    @property
    def operation(self) -> GroupOperation:
        """Group operation."""
        return self._operation


class SectionBlock(DiagramObject):  # numpydoc ignore=PR01
    """Section block definition:

    - *object* ::= ( *description* [[ *local_objects* ]] )
    - *description* ::= *scope_section*

    """

    def __init__(
        self,
        section: scopes.ScopeSection,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(locals=locals, pragmas=pragmas)
        self._section = section
        common.SwanItem.set_owner(self, section)

    @property
    def section(self) -> scopes.ScopeSection:
        """Section object of diagram object."""
        return self._section

    @property
    def sources(self) -> List[tuple["DiagramObject", Optional[GroupAdaptation]]]:
        """This method must not be called for a SectionBlock"""
        raise ScadeOneException("SectionBlock.sources() call")

    @property
    def targets(self) -> List[tuple["DiagramObject", Optional[GroupAdaptation]]]:
        """This method must not be called for a SectionBlock"""
        raise ScadeOneException("SectionBlock.targets() call")


class DefByCaseBlockBase(DiagramObject):  # numpydoc ignore=PR01
    """Def-by-case graphical definition (automaton or activate if/when):

    - *object* ::= ( *description* [[ *local_objects* ]] )
    - *description* ::= *def_by_case*

    This class is a base class for StateMachineBlock, ActivateIfBlock and ActivateWhenBlock
    and is used as a proxy to the internal DefByCase object.

    """

    def __init__(
        self,
        def_by_case: DefByCase,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(locals=locals, pragmas=pragmas)
        self._def_by_case = def_by_case
        common.SwanItem.set_owner(self, def_by_case)

    @property
    def def_by_case(self) -> DefByCase:
        """Def-by-case object."""
        return self._def_by_case

    def __getattr__(self, name: str):
        """Proxy to the DefByCase object."""
        return getattr(self._def_by_case, name)

    @property
    def sources(self) -> List[tuple["DiagramObject", Optional[GroupAdaptation]]]:
        """This method must not be called for a *def-by-case* block."""
        raise ScadeOneException("DefByCaseBlockBase.sources() call")

    @property
    def targets(self) -> List[tuple["DiagramObject", Optional[GroupAdaptation]]]:
        """This method must not be called for a *def-by-case* block."""
        raise ScadeOneException("DefByCaseBlockBase.targets() call")


class StateMachineBlock(DefByCaseBlockBase):  # numpydoc ignore=PR01
    """State machine block definition:

    - *object* ::= ( *description* [[ *local_objects* ]] )
    - *description* ::= *[lhs :] state_machine*

    A *StateMachineBlock* is a proxy to the internal :py:class:`StateMachine` object, therefore
    the methods and properties of the *StateMachine* object can be accessed directly.
    """

    def __init__(
        self,
        def_by_case: StateMachine,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(def_by_case, locals, pragmas)

    @property
    def state_machine(self) -> StateMachine:
        """State machine object."""
        return self.def_by_case


class ActivateIfBlock(DefByCaseBlockBase):  # numpydoc ignore=PR01
    """Activate-if block definition:

    - *object* ::= ( *description* [[ *local_objects* ]] )
    - *description* ::= *[lhs :] activate [[ luid ]] if_activation*

    A *ActivateIF* is a proxy to the internal :py:class:`ActivateIf` object, therefore
    the methods and properties of the *ActivateIf* object can be accessed directly.

    """

    def __init__(
        self,
        def_by_case: ActivateIf,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(def_by_case, locals, pragmas)

    @property
    def activate_if(self) -> ActivateIf:
        """Activate if object."""
        return self.def_by_case


class ActivateWhenBlock(DefByCaseBlockBase):  # numpydoc ignore=PR01
    """Activate-when block definition:

    - *object* ::= ( *description* [[ *local_objects* ]] )
    - *description* ::= *[lhs :] activate [[ luid ]] when_activation*

    A *ActivateIF* is a proxy to the internal :py:class:`ActivateWhen` object, therefore
    the methods and properties of the *ActivateIf* object can be accessed directly.

    """

    def __init__(
        self,
        def_by_case: ActivateWhen,
        locals: Optional[List[DiagramObject]] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        super().__init__(def_by_case, locals, pragmas)

    @property
    def activate_when(self) -> ActivateWhen:
        """Activate when object."""
        return self.def_by_case


# Diagram navigation
# ------------------------------------------------------------


class DiagramNavigation:
    """Class handling navigation through Diagram objects.

    Parameters
    ----------
    diagram: Diagram
        Diagram object to navigate.
    """

    def __init__(self, diagram: Diagram) -> None:
        self._block_table = {}
        self._wires_of_target = defaultdict(list)
        self._wires_of_source = defaultdict(list)
        self._diagram = diagram

    def get_block(self, lunum: common.Lunum) -> Block:
        """Getting specific block."""
        return self._block_table[lunum.value]

    def with_source(self, lunum: common.Lunum) -> List[Wire]:
        """Returning list of wires that have a specific source."""
        return self._wires_of_source[lunum.value]

    def with_target(self, lunum: common.Lunum) -> List[Wire]:
        """Returning list of wires that have a specific target."""
        return self._wires_of_target[lunum.value]

    def get_wire_source(
        self, wire: Wire
    ) -> tuple[
        DiagramObject,
        GroupAdaptation,
        Union[List[tuple[DiagramObject, GroupAdaptation]], None],
    ]:
        """Get source block and adaptations of a wire. Also get the list of target blocks and adaptations."""
        from_block = self.get_block(wire.source.port.lunum)
        from_block_adaptation = wire.source.adaptation
        to_blocks = []
        for target in wire.targets:
            to_block = self.get_block(target.port.lunum)
            to_blocks.append((to_block, target.adaptation))
        return from_block, from_block_adaptation, to_blocks if to_blocks else None

    def get_wire_targets(
        self, wire: Wire
    ) -> List[tuple[DiagramObject, GroupAdaptation, GroupAdaptation]]:
        """Getting a list of targets block and adaptations of a wire."""
        list_targets = []
        from_block_adaptation = wire.source.adaptation
        for target in wire.targets:
            block = self.get_block(target.port.lunum)
            to_block_adaptation = target.adaptation
            list_targets.append((block, from_block_adaptation, to_block_adaptation))
        return list_targets

    def get_block_sources(
        self, obj: DiagramObject
    ) -> List[tuple[DiagramObject, GroupAdaptation, Union[List[GroupAdaptation], None]]]:
        """A block sources list of a Diagram Object."""
        if len(obj.locals) != 0:
            locals = [local.lunum for local in obj.locals]
            target_wires = []
            for lunum in locals:
                target_wires.extend(self.with_target(lunum))
        else:
            target_wires = self.with_target(obj.lunum)
        sources = [self.get_wire_source(wire) for wire in target_wires]
        return self._filter_sources(obj, sources)

    def _filter_sources(
        self,
        obj: DiagramObject,
        sources: List[
            tuple[
                DiagramObject,
                GroupAdaptation,
                Optional[tuple[DiagramObject, GroupAdaptation]],
            ]
        ],
    ) -> List[tuple[DiagramObject, GroupAdaptation, Union[List[GroupAdaptation], None]]]:
        """Filter sources (object, adaptation) of a given diagram object."""
        sources_dict = {}
        for source in sources:
            target_obj_adps = self._filter_target_blocks(source[2], obj)
            for target_obj_adp in target_obj_adps:
                # Stock a source (object, adaptation) and the corresponding target adaptations
                source_obj_adp = (source[0], source[1])
                if not sources_dict.get(source_obj_adp):
                    sources_dict[source_obj_adp] = {}
                    if target_obj_adp[1]:
                        sources_dict[source_obj_adp] = {target_obj_adp[1]}
                elif target_obj_adp[1]:
                    sources_dict[source_obj_adp].add(target_obj_adp[1])
        obj_sources = []
        # Convert the (source, target adaptations) dictionary
        # to a list of tuples: (source_block, source_adaptation, targets)
        for source_obj_adp, target_adps in sources_dict.items():
            obj_sources.append(
                (
                    source_obj_adp[0],
                    source_obj_adp[1],
                    list(target_adps) if target_adps else None,
                )
            )
        return obj_sources

    @staticmethod
    def _filter_target_blocks(
        target_obj_adps: Optional[List[tuple[DiagramObject, GroupAdaptation]]], obj
    ) -> List[tuple[DiagramObject, GroupAdaptation]]:
        """Filter target objects and adaptations of a given diagram object."""
        if not target_obj_adps:
            return []
        filtered_target_obj_adps = []
        for target_obj_adp in target_obj_adps:
            if target_obj_adp[0].locals:
                for local in target_obj_adp[0].locals:
                    if local.lunum == obj.lunum:
                        filtered_target_obj_adps.append((local, None))
            if target_obj_adp[0].lunum == obj.lunum:
                filtered_target_obj_adps.append(target_obj_adp)
        return filtered_target_obj_adps

    def get_block_targets(
        self, obj: DiagramObject
    ) -> List[tuple[DiagramObject, GroupAdaptation, GroupAdaptation]]:
        """A list of targets block of a Diagram Object."""
        lunum = obj.lunum
        if not isinstance(obj.owner, Diagram):
            lunum = obj.owner.lunum
        if not lunum:
            raise ScadeOneException(
                "Cannot get targets of a block without a locally unique number (LUNUM)"
            )
        targets = []
        for wire in self.with_source(lunum):
            targets.extend(self.get_wire_targets(wire))
        return targets

    def consolidate(self):
        """Retrieve wire sources, wire targets and blocks from the Diagram Object."""

        def explore_object(obj: DiagramObject):
            if isinstance(obj, (SectionBlock, DefByCaseBlockBase)):
                return
            if isinstance(obj, Wire):
                # process targets
                # _wire_of_target: table which stores wires from
                # target block found in wire
                wire = cast(Wire, obj)
                for target in wire.targets:
                    if not target.is_connected:
                        continue
                    if target.port.is_self:
                        continue
                    self._wires_of_target[target.port.lunum.value].append(wire)
                # process source
                # _wire_of_source: table which stores wires from
                # source block found in wire
                if wire.source.is_connected and not wire.source.port.is_self:
                    self._wires_of_source[wire.source.port.lunum.value].append(wire)
            else:
                lunum = obj.lunum
                if lunum is None:
                    return
                self._block_table[lunum.value] = obj
                for local in obj.locals:
                    self._block_table[local.lunum.value] = obj

        for obj in self._diagram.objects:
            explore_object(obj)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import logging
from pathlib import Path
import sys

from ansys.scadeone.core import ScadeOne, ScadeOneException, full_version
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.common.versioning import FormatVersions
from ansys.scadeone.core.svc.fmu import FMU_2_Export

# cSpell: ignore outdir, oper


def show_formats() -> bool:
    # Show supported formats: --formats option
    print(FormatVersions.get_versions())
    return True


def fmu_export_command(args) -> bool:
    print(
        f"Build FMU package for job {args.job_name} under directory {args.outdir}",
        flush=True,
    )

    try:
        app = ScadeOne(args.install_dir)
        project = app.load_project(args.project)
        fmu = FMU_2_Export(project, args.job_name, args.oper_name, args.max_variables)

        outdir = f"FMU_Export_{args.kind}_{args.job_name}" if args.outdir is None else args.outdir
        if args.period is None:
            fmu.generate(args.kind, outdir)
        else:
            fmu.generate(args.kind, outdir, args.period)

        if args.build_arguments is not None:
            build_args = dict(args.build_arguments)
            if build_args.get("user_sources", "") != "":
                build_args["user_sources"] = build_args["user_sources"].split(",")
            if build_args.get("cc_opts", "") != "":
                build_args["cc_opts"] = build_args["cc_opts"].split(",")
            if build_args.get("_opts", "") != "":
                build_args["cc_opts"] = build_args["cc_opts"].split(",")
        else:
            build_args = None
        fmu.build(args.with_sources, build_args)
        return True
    except ScadeOneException as error:
        print("ERROR -", error.args[0], file=sys.stderr)
        return False


def python_wrapper_command(args) -> bool:
    from ansys.scadeone.core.svc.wrapper.python_wrapper import PythonWrapper

    print(
        f"Generate Python wrapper for project {args.project}",
        flush=True,
    )

    try:
        app = ScadeOne(args.install_dir)
        project_path = Path(args.project)
        if not project_path.is_absolute():
            project_path = Path.cwd() / project_path
        project = app.load_project(project_path)
        job_name = args.job
        out_name = None
        if args.output:
            out_name = args.output
        target_path = None
        if args.target_dir:
            target_path = args.target_dir

        py_wrapper = PythonWrapper(project, job_name, out_name, target_path)
        py_wrapper.generate()

        print(
            f"Files generated under {py_wrapper._target_dir()}",
            flush=True,
        )
        return True
    except ScadeOneException as error:
        print("ERROR -", error.args[0], file=sys.stderr)
        return False


def view_sd_command(args) -> bool:
    import ansys.scadeone.core.svc.simdata as sd

    if args.sd:
        sd_path = Path(args.sd)
        if not sd_path.exists():
            print(f"File not found: {sd_path}")
            return False
        sd_fd = sd.open_file(str(sd_path))
        print(sd_fd)
        sd_fd.close()
        return True
    return False


def main() -> bool:
    """Scade One Python command line.

    Returns True if the command line is executed successfully."""
    parser = argparse.ArgumentParser(
        prog="pyscadeone",
        description="Scade One Python library command line tool",
        epilog="For more information see: "
        "https://www.ansys.com/products/embedded-software/ansys-scade-one",
        fromfile_prefix_chars="@",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=full_version,
        help="%(prog)s version",
    )
    parser.add_argument(
        "--formats",
        action="store_true",
        help="Shows supported formats",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-v",
        "--verbosity",
        action="count",
        default=0,
        help="Activates verbose mode. Several occurrences increase verbosity level",
    )

    subparser = parser.add_subparsers(
        title="Sub-commands",
        help="Use: command --help for help (ex: pyscadeone fmu --help)",
        dest="subparser_command",
    )

    # FMU export command
    fmu_parser = subparser.add_parser("fmu", help="Generate FMU")
    fmu_parser.add_argument(
        "project",
        type=Path,
        help="Scade One project",
    )
    fmu_parser.add_argument(
        "job_name",
        type=str,
        help="Generated Code job name",
    )
    fmu_parser.add_argument(
        "-inst",
        "--install_dir",
        type=Path,
        help="Scade One installation directory",
    )
    fmu_parser.add_argument(
        "-op",
        "--oper_name",
        type=str,
        default="",
        help="Root operator name",
    )
    fmu_parser.add_argument(
        "-max",
        "--max_variables",
        type=int,
        default=1000,
        help="Maximum number on FMI variables (flattened sensors, inputs and "
        "outputs) supported by the export (1000 by default).",
    )
    fmu_parser.add_argument(
        "-k",
        "--kind",
        type=str,
        default="ME",
        help="FMI kind: 'ME' for Model Exchange (default), 'CS' for Co-Simulation",
    )
    fmu_parser.add_argument(
        "-o",
        "--outdir",
        type=Path,
        help="Directory where the FMU is built (by default, 'FMU_Export_<kind>_<job_name>')",
    )
    fmu_parser.add_argument(
        "-p",
        "--period",
        type=float,
        help="Execution period in seconds",
    )
    fmu_parser.add_argument(
        "-ws",
        "--with_sources",
        action="store_true",
        help="Keep the sources in the FMU package",
    )
    #    fmu_parser.add_argument("-args", "--args", type=str, help="build arguments")
    fmu_parser.add_argument(
        "-args",
        "--build_arguments",
        type=str,
        action="append",
        nargs=2,
        metavar=("key", "value"),
        help=(
            """\
            Build arguments. Use one -args argument per key. Supported keys are:
            cc: compiler name (only gcc supported),
            arch: compiler architecture (only win64 supported),
            gcc_path: path on the bin directory where gcc is located,
            user_sources: list (comma separated) of user source files and directories
            (code, includes),
            cc_opts: list (comma separated) of extra compiler options,
            link_opt: list (comma separated) of extra link (dll creation) options,
            swan_config_begin: data to insert at the beginning of swan_config.h,
            swan_config_end: data to insert at the end of swan_config.h."""
        ),
    )
    fmu_parser.set_defaults(func=fmu_export_command)

    # Python wrapper command
    pw_parser = subparser.add_parser("pycodewrap", help="Generates Scade One wrapper in Python")
    pw_parser.add_argument(
        "--install-dir",
        type=Path,
        help="Scade One installation directory",
        required=True,
    )
    pw_parser.add_argument(
        "project",
        type=Path,
        help="Scade One project",
    )
    pw_parser.add_argument("-j", "--job", type=str, help="Generated Code job name", required=True)
    pw_parser.add_argument(
        "-o",
        "--output",
        help="Name of the Python wrapper module. By default, the name is `root_wrapper`",
    )
    pw_parser.add_argument(
        "--target-dir",
        type=Path,
        help="Target wrapper directory. By default, a directory with the wrapper name is created "
        "in the current directory.",
    )

    pw_parser.set_defaults(func=python_wrapper_command)

    # Simulation data
    sd_parser = subparser.add_parser("simdata", help="Simulation data related command")
    sd_parser.add_argument(
        "--show",
        help="Show content of a simulation data file",
        dest="sd",
    )
    sd_parser.set_defaults(func=view_sd_command)

    # Parsing
    parser_args = parser.parse_args()

    if parser_args.verbosity > 0:
        if parser_args.verbosity == 1:
            level = logging.INFO
        else:
            level = logging.DEBUG
        for handler in LOGGER.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(level)

    LOGGER.debug(f"Running pyscadeone command line with: {parser_args}")

    ret = False
    if parser_args.formats:
        ret = show_formats()
    elif parser_args.subparser_command:
        ret = parser_args.func(parser_args)
    sys.exit(0 if ret else 1)


if __name__ == "__main__":
    sys.exit(0 if main() else 1)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import shutil
from collections import namedtuple
from keyword import iskeyword
from pathlib import Path
import re
from typing import List, Optional, Tuple, Union

import jinja2

from ansys.scadeone.core import ScadeOneException
from ansys.scadeone.core.common.versioning import FormatVersions
from ansys.scadeone.core.project import Project
from ansys.scadeone.core.svc.build_system import (
    BuildConfig,
    BuildSystem,
    Target,
    TargetKind,
)
import ansys.scadeone.core.svc.generated_code as gc

# flake8: noqa

PredefInfo = namedtuple("PredefInfo", ["type_name", "init_value"])

predefs_ctypes = {
    "int8": PredefInfo("ctypes.c_int8", "0"),
    "int16": PredefInfo("ctypes.c_int16", "0"),
    "int32": PredefInfo("ctypes.c_int32", "0"),
    "int64": PredefInfo("ctypes.c_int64", "0"),
    "uint8": PredefInfo("ctypes.c_uint8", "0"),
    "uint16": PredefInfo("ctypes.c_uint16", "0"),
    "uint32": PredefInfo("ctypes.c_uint32", "0"),
    "uint64": PredefInfo("ctypes.c_uint64", "0"),
    "bool": PredefInfo("ctypes.c_uint8", "0"),
    "char": PredefInfo("ctypes.c_int8", "0"),
    "float32": PredefInfo("ctypes.c_float", "0.0"),
    "float64": PredefInfo("ctypes.c_double", "0.0"),
    "size": PredefInfo("ctypes.c_int64", "0"),
}

predefs_native = {
    "int8": PredefInfo("int", "0"),
    "int16": PredefInfo("int", "0"),
    "int32": PredefInfo("int", "0"),
    "int64": PredefInfo("int", "0"),
    "uint8": PredefInfo("int", "0"),
    "uint16": PredefInfo("int", "0"),
    "uint32": PredefInfo("int", "0"),
    "uint64": PredefInfo("int", "0"),
    "bool": PredefInfo("bool", "False"),
    "char": PredefInfo("int", "0"),
    "float32": PredefInfo("float", "0.0"),
    "float64": PredefInfo("float", "0.0"),
    "size": PredefInfo("int", "0"),
}

predefs_values = {"false": "0", "true": "1"}


class PythonWrapper:
    def __init__(
        self,
        project: Project,
        job: str,
        output: Optional[str] = None,
        target_dir: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Generate a Scade One wrapper in Python.

        Parameters
        ----------
        project : Project
            Scade One project.

        job : str
            Swan Code Generation job name.

        output : str
            Wrapper name. If not provided, the name is "root_wrapper".

        target_dir : str, Path
            Wrapper location. If not provided, the wrapper is generated in a directory named with the wrapper name
            located in the current directory.
        """
        self._banner = "{0} {1}".format(
            FormatVersions.description("wrapper"), FormatVersions.version("wrapper")
        )
        self._project = project
        self._code_gen = gc.GeneratedCode(project, job)
        self._model_ops = None
        self._out_name = output or "root_wrapper"
        self._target_path = target_dir

    def _init_data(self):
        """
        Initialize data
        """

        # Get all model operators
        _opts = self._code_gen.root_operators
        self._model_ops = [self._code_gen.get_model_operator(_opt) for _opt in _opts]
        # Set output name
        if not self._out_name:
            self._out_name = "_".join([_nm.split("::")[-1] for _nm in _opts])

    def _find_elaboration(self) -> Tuple[bool, str]:
        """
        Find a "elaboration" is present in the mapping file
        Returns
        -------
        Tuple[bool, str]
            If "elaboration" is present, returns True and its value, otherwise, returns False and empty.
        """

        _ela = self._code_gen.get_elaboration_function()
        return (True, _ela.name) if _ela else (False, "")

    def _target_dir(self) -> Path:
        """
        Get a target dir

        Returns
        -------
        Path
            Return a Path
        """
        target_path = (
            Path(self._target_path) / self._out_name
            if self._target_path
            else Path.cwd() / self._out_name
        )
        if not target_path.is_absolute():
            target_path = Path.cwd() / target_path
        target_path.mkdir(parents=True, exist_ok=True)
        return target_path

    def _get_predef_info(self, c_type_name: str, native: bool) -> PredefInfo:
        if native:
            return predefs_native[c_type_name] if c_type_name in predefs_native else None
        else:
            return predefs_ctypes[c_type_name] if c_type_name in predefs_ctypes else None

    def _get_python_name(self, name: str) -> str:
        return name + "_" if iskeyword(name) else name

    def _get_python_type_name(self, type_base: gc.TypeBase, native: bool = False):
        """
        Get Python type name

        Parameters
        ----------
        type_base : TypeBase
            An instance of TypeBase
        native : bool, optional
            A type native option, by default False
        sizes : _type_, optional
            An array size, by default None

        Returns
        -------
        str
            Returns type name
        """

        if type_base and (type_base.is_scalar or type_base.is_enum_type):
            pi = self._get_predef_info(type_base.m_name, native)
            assert pi, f"No predef info was found for predef type {type_base.m_name}"
            name = pi.type_name
            return self._get_python_name(name)
        elif type_base and type_base.is_structure:
            return type_base.path
        elif type_base and type_base.is_array:
            pi = self._get_predef_info(type_base.m_name, False)
            size = " * ".join([str(s) for s in type_base.sizes])
            return f"{pi.type_name} * {size}"
        else:
            # TODO with others
            return ""

    def _get_initial_value(self, type_base: gc.TypeBase, native: bool = False) -> str:
        """
        Get default value corresponding to a given type

        Parameters
        ----------
        type_base : TypeBase
            An instance of TypeBase
        native : bool, optional
            A type native option, by default False

        Returns
        -------
        str
            Returns type name
        """

        if type_base and (type_base.is_scalar or type_base.is_enum_type):
            pi = self._get_predef_info(type_base.m_name, native)
            return pi.init_value
        elif type_base and type_base.is_array:
            return "()"
        elif type_base and type_base.is_structure:
            return type_base.path + "()"
        else:
            return ""

    def _get_enum_name(self, type_base: gc.TypeBase) -> str:
        if type_base.is_enum_type:
            return type_base.enum_name
        return ""

    def _get_array_info(self, type_base: gc.TypeBase) -> str:
        if type_base.is_array:
            return "tuple"
        return ""

    def _is_array_types(self, io_data: list) -> bool:
        for _elm in io_data:
            if any(_in[4] for _in in _elm["inputs"]["parameters"]):
                return True
            if any(_out[5] for _out in _elm["outputs"]["parameters"]):
                return True
        return False

    def _get_operator_ios(
        self, model_op: gc.ModelOperator, _mapping: dict
    ) -> tuple[List[tuple[str]], List[tuple[str]]]:
        """
        Get inputs/outputs of an operator

        Parameters
        ----------
        _mapping : dict
            A dict of type mapping data

        Returns
        -------
        tuple[List[tuple[str]], List[tuple[str]]]
            A tuple of inputs/outputs
        """
        _inputs = []
        _outputs = []
        _io_dct = {}
        _ou_dct = {}

        for _elm in self._code_gen.get_field_value(
            self._code_gen.get_code_elem(model_op._id), "parameters"
        ):
            _io_dct[_elm["id"]] = _elm["type"]

        for _elm in [_in for _in in model_op.inputs]:
            if _elm._id in _io_dct:
                _typ = self._code_gen.get_type(_mapping[_io_dct[_elm._id]], _io_dct[_elm._id])
                _nm = self._get_python_type_name(_typ, True)
                _val = self._get_initial_value(_typ, True)
                _enum = self._get_enum_name(_typ)
                _array = self._get_array_info(_typ)
                _inputs.append((_elm.name, _nm, _val, _enum, _array))
        for _elm in self._code_gen.get_interface_file(model_op._id)["declarations"]:
            if _elm[0] == "struct":
                for _id in _elm[1]["fields"]:
                    _ou_dct[_id["id"]] = _id["type"]
        for _elm in [_ot for _ot in model_op.outputs]:
            if _elm._id in _ou_dct:
                _rtn = _ou_dct[_elm._id]
            else:
                _chk = self._code_gen.get_field_value(
                    self._code_gen.get_code_elem(model_op._id), "return_type"
                )
                if _chk:
                    # If "return_type" is available in the mapping file
                    _rtn = _chk
                else:
                    _rtn = _io_dct[_elm._id]
            _typ = self._code_gen.get_type(_mapping[_rtn], _rtn)
            _nm = self._get_python_type_name(_typ, True)
            _val = self._get_initial_value(_typ, True)
            _enum = self._get_enum_name(_typ)
            _array = self._get_array_info(_typ)
            _cty = self._get_python_type_name(_typ, False)
            _outputs.append((_elm.name, _nm, _val, _enum, _array, _cty))
        return _inputs, _outputs

    def _generate_ios_context(
        self, model_op: gc.ModelOperator, context_data: List[str], _mapping: dict
    ) -> List[str]:
        """
        Generate inputs/outputs context

        Parameters
        ----------
        context_data : List[str]
            _description_
        _mapping : dict
            _description_

        Returns
        -------
        List[str]
            Return a list of contexts
        """
        _rtn = []
        for _idx, _elm in enumerate(context_data):
            if (
                _elm["name"] not in [_elm.name for _elm in model_op.inputs]
                and self._code_gen.get_code_id(_elm["id"], "ContextType", True) == -1
            ):
                _typ = self._code_gen.get_type(_mapping[_elm["type"]], _elm["type"])
                _nm = self._get_python_type_name(_typ)
                if isinstance(_typ, gc.Array):
                    _nm = f"({_nm})"
                _rtn.append(f"self._{_elm['name']} = {_nm}.from_address(context + offsets[{_idx}])")
            elif _elm.get("pointer"):
                _typ = self._code_gen.get_type(_mapping[_elm["type"]], _elm["type"])
                _nm = self._get_python_type_name(_typ)
                _rtn.append(f"self._{_elm['name']} = {_nm}()")
        return _rtn

    def _generate_cycle_parameters(
        self, model_op: gc.ModelOperator, parameters: dict, _mapping: dict
    ) -> tuple[str, List[str]]:
        """
        Generate cycle parameters

        Parameters
        ----------
        parameters : dict
            A dict of parameters data
        _mapping : dict
            A dict of type data

        Returns
        -------
        tuple[List[str], List[str]]
            Return a tuple of parameters and parameter names
        """
        _param = []
        _params_name = []
        for _fld in parameters:
            if _fld["type"] in _mapping:
                if self._code_gen.get_role_type(model_op._id, _fld["type"]) == "ContextType":
                    _rtn = "ctypes.c_void_p"
                else:
                    _rtn = self._get_python_type_name(
                        self._code_gen.get_type(_mapping[_fld["type"]], _fld["type"])
                    )
                if _fld.get("pointer"):
                    _rtn = f"ctypes.POINTER({_rtn})"
                _param.append(_rtn)
        for _fld in parameters:
            if _fld["type"] in _mapping:
                if self._code_gen.get_role_type(model_op._id, _fld["type"]) == "ContextType":
                    _params_name.append("out_c")
                else:
                    if _fld["name"] in [_it.name for _it in model_op.inputs]:
                        _params_name.append(_fld["name"])
        return _param, _params_name

    def _generate_method_data(
        self, model_op: gc.ModelOperator, lst_method: List[str], save: dict
    ) -> None:
        """
        Generate method data

        Parameters
        ----------
        lst_method : List[str]
            A list of methods
        save : dict
            An output
        """
        for _role in lst_method:
            _name = "_".join([x.lower() for x in re.findall("[A-Z][^A-Z]*", _role)])
            _dta = self._code_gen.get_method_data(model_op._id, _role)
            save[_name] = _dta

    def _generate_def(self, def_pathname: Path) -> None:
        """
        Generate .def file for exported symbols.

        Parameters
        ----------
        def_pathname : Path
            Path of the .def file to generate
        """

        # Initialize exports data
        _exp = {}
        _id = 1
        for _model_op in self._model_ops:
            # Add cycle method data
            if _model_op.cycle_method:
                _exp[_model_op.cycle_method.name] = _id
                _id = _id + 1
            # Add init method data
            if _model_op.init_method:
                _exp[_model_op.init_method.name] = _id
                _id = _id + 1
            # Add reset method data
            if _model_op.reset_method:
                _exp[_model_op.reset_method.name] = _id
                _id = _id + 1
        # Add elaboration data
        _fnd, _elaboration = self._find_elaboration()
        if _fnd:
            _exp[_elaboration] = _id
        # Add sensors data
        for _code in self._code_gen.code:
            if _code["interface_file"] != "swan_sensors.h":
                continue
            for _decl in _code["declarations"]:
                if _decl[0] != "global":
                    continue
                _exp[_decl[1]["name"]] = _id
                _id = _id + 1
        # Create *.def file
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("exports_wrapper.j2")

        content = template.render(TOOL_WRAPPER=self._banner, EXP_DEF=_exp)

        def_pathname.write_text(content)

    def _generate_c_code(self, c_pathname: Path) -> None:
        """
        Generate *.c file

        Parameters
        ----------
        c_pathname : Path
            Path of the .c file to generate
        """

        _data = []
        _sns_data = {}
        for _model_op in self._model_ops:
            # Initialize operator data
            _op_dict = {}
            _op_dict["ios_declared"] = []
            _op_dict["init_c_name"] = ""
            _op_dict["reset_parameters"] = []
            _op_dict["opt_name"] = self._code_gen.get_operator_name(_model_op._id)
            _op_dict["opt_path"] = _model_op.path

            # Update the data for method
            self._generate_method_data(
                _model_op, ["InitMethod", "ResetMethod", "ContextType"], _op_dict
            )
            if _op_dict["init_method"]:
                _op_dict["init_c_name"] = self._code_gen.get_field_value(
                    _op_dict["init_method"], "name"
                )

            if _op_dict["reset_method"]:
                _op_dict["reset_parameters"] = self._code_gen.get_field_value(
                    _op_dict["reset_method"], "parameters"
                )

            if _op_dict["context_type"] and _model_op.cycle_method:
                _ios_declared = []
                for _fld in self._code_gen.get_field_value(_op_dict["context_type"], "fields"):
                    if _fld["id"] in self._code_gen.get_model_id_mapping():
                        _ios_declared.append(_fld["name"])
                _op_dict["ios_declared"] = _ios_declared
            _data.append(_op_dict)

        # Sensor data
        _type_mapping = self._code_gen.get_type_id_mapping()
        for _elm in self._code_gen.code:
            _inc = _elm.get("interface_file")
            if "sensor" in _inc:
                _sns_data["include"] = str(_inc)
                break
        _sns = self._code_gen.get_model_elements("sensor")
        _sns_data["elements"] = []
        for _, _elm in _sns:
            _dt = {}
            _dt["path"] = _elm["path"]
            _, _vl = _type_mapping[_elm["id"]]
            _dt["name"] = _vl["name"]
            _typ = self._code_gen.get_type(_type_mapping[_vl["type"]], _vl["type"])
            _dt["type"] = _typ.path
            _sns_data["elements"].append(_dt)

        # Create *.c code file
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("C_wrapper.j2")

        content = template.render(
            TOOL_WRAPPER=self._banner, OPERATORS_DATA=_data, SENSORS=_sns_data
        )

        c_pathname.write_text(content)

    def _generate_python(self, py_pathname: Path) -> None:
        """
        Generate *.py file

        Parameters
        ----------
        py_pathname : Path
            Path of the .py file to generate
        """

        _data = []
        _enum_data = {}
        _struct_data = {}
        _sns_data = []

        for _model_op in self._model_ops:
            # Initialize operator data
            _op_dict = {}
            _op_dict["ios_context"] = []
            _op_dict["op_name"] = self._code_gen.get_operator_name(_model_op._id)
            _op_dict["inputs"] = {"name": "Inputs", "parameters": []}
            _op_dict["outputs"] = {"name": "Outputs", "parameters": []}

            # Update the data for method
            _type_mapping = self._code_gen.get_type_id_mapping()
            _op_dict["inputs"]["parameters"], _op_dict["outputs"]["parameters"] = (
                self._get_operator_ios(_model_op, _type_mapping)
            )

            _op_dict["op_context"] = {}
            _ctx = []
            self._generate_method_data(
                _model_op, ["ResetMethod", "ContextType", "CycleMethod"], _op_dict
            )
            if _op_dict["context_type"] and _model_op.cycle_method:
                _op_dict["op_context"]["c_name"] = self._code_gen.get_operator_name(_model_op._id)
                _, _out = self._code_gen.get_function_ios(_model_op._id, "ContextType")
                _op_dict["op_context"]["py_name"] = "_out_c" if _out else ""
                for _fld in self._code_gen.get_field_value(_op_dict["context_type"], "fields"):
                    if _fld["id"] in self._code_gen.get_model_id_mapping():
                        _ctx.append(_fld)
                _op_dict["op_context"]["len_context"] = len(_ctx)

            if _op_dict["reset_method"]:
                _op_dict["op_reset"] = {
                    "c_name": self._code_gen.get_field_value(_op_dict["reset_method"], "name"),
                    "parameters": self._code_gen.get_field_value(
                        _op_dict["reset_method"], "parameters"
                    ),
                }
            else:
                _op_dict["op_reset"] = {}

            _op_dict["ios_context"] = self._generate_ios_context(_model_op, _ctx, _type_mapping)
            _rtn = None
            cycle_method_content = _op_dict["cycle_method"]["content"]
            if cycle_method_content.get("return_type"):
                _rtn = _op_dict["outputs"]["parameters"][0]
            _op_dict["op_cycle"] = {
                "c_name": self._code_gen.get_field_value(_op_dict["cycle_method"], "name"),
                "parameter": "",
                "parameters_name": [],
                "return": _rtn,
            }
            (
                _op_dict["op_cycle"]["parameter"],
                _op_dict["op_cycle"]["parameters_name"],
            ) = self._generate_cycle_parameters(
                _model_op,
                self._code_gen.get_field_value(_op_dict["cycle_method"], "parameters"),
                _type_mapping,
            )

            _data.append(_op_dict)

        # Enumeration / Structure types
        _type_mapping = self._code_gen.get_type_id_mapping()
        _code_mapping = self._code_gen.get_model_id_mapping()
        _links = self._code_gen._get_links()

        for _elm in _type_mapping.values():
            if _elm[0] == "enum":
                enum_values = []
                for value in _elm[1]["values"]:
                    enum_values.append(value["name"])
                _enum_data[_elm[1]["name"]] = enum_values
            elif _elm[0] == "struct":
                struct_name = _elm[1].get("name")
                struct_values = []
                if _elm[1].get("id") in _code_mapping.keys():
                    for value in _elm[1]["fields"]:
                        id_type = value["type"]
                        id_to_map = _type_mapping[id_type][1]["id"]
                        id_type = _links.get(id_to_map)
                        field_type = "None"  # TODO change for complex structures
                        if id_type:
                            field_type_name = _code_mapping.get(id_type)[1]["name"]
                            field_type = self._get_predef_info(field_type_name, False)[0]
                        struct_values.append([value["name"], field_type])
                    _struct_data[struct_name] = struct_values

        # Sensor data
        _sns = self._code_gen.get_model_elements("sensor")
        for _, _elm in _sns:
            _, _vl = _type_mapping[_elm["id"]]
            _snm = _elm["path"].split("::")[1]
            _spt = _vl["name"]
            _typ = self._code_gen.get_type(_type_mapping[_vl["type"]], _vl["type"])
            _cty = self._get_python_type_name(_typ)
            _pty = self._get_python_type_name(_typ, True)
            _sns_data.append((_snm, _cty, _spt, _pty))

        # Elaboration data
        _, _ela = self._find_elaboration()

        # Create *.py code file
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("Py_wrapper.j2")

        content = template.render(
            TOOL_WRAPPER=self._banner,
            OPERATORS_DATA=_data,
            ELABORATION=_ela,
            SENSORS=_sns_data,
            ENUM_TYPES=_enum_data,
            STRUCTURE_TYPES=_struct_data,
            IS_ARRAY_TYPES=self._is_array_types(_data),
        )

        py_pathname.write_text(content)

    def _generate_swan_config(self, config_pathname: Path) -> None:
        """
        Generate swan_config.h file

        Parameters
        ----------
        config_pathname : Path
            Path of the .h file to generate
        """
        includes = ""
        resources = []

        # TODO: to verify this code works well with others User Stories related to Types and Imported Code
        # Get all import and  *.h header files defined for the project
        _data = self._project.data
        _path = self._project.directory
        for resource in _data["Resources"]:
            if resource["Kind"] in ("ImportedTypes", "HeaderFile"):
                new_resource = _path.parent.joinpath(resource["Path"]).resolve()
                if new_resource not in resources:
                    resources.append(new_resource)

        includes = "\n" + "\n".join(f'#include "{r}"' for r in resources) + "\n"

        # Create swan_config.h
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(Path(__file__).parent.parent / "common/templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("swan_config_template.h")

        content = template.render(HOOK_BEGIN="", HOOK_END="")

        swan_config = content.replace("#include <string.h>", "#include <string.h>" + includes)
        config_pathname.write_text(swan_config)

    def _generate_declaration_files(self) -> List[str]:
        """
        Get all files generated

        Returns
        -------
        List[str]
            List of files generated
        """

        _lt = []
        for _elm in self._code_gen.code:
            if _elm.get("implementation_file"):
                _lt.append(_elm["implementation_file"])
            if _elm.get("interface_file"):
                _lt.append(_elm["interface_file"])
        return _lt

    def _build_wrapper(self):
        """
        Build wrapper using the C files generated by Swan CG
        and the C and DEF files generated by the wrapper. The temporary files are generated
        in a temporary directory called "build",
        located in the same wrapper directory. If the build is successful, the temporary directory
        is removed, otherwise, the directory is kept for debugging purposes.
        The builder uses the BuildSystem located in the Scade One installation.
        """
        cfg = BuildConfig()
        build_dir = Path(self._target_dir()) / "build"
        build_dir.mkdir(parents=True, exist_ok=True)
        cfg.working_dir = str(build_dir)
        cfg.c_files = [str(file) for file in self._target_dir().glob("*.c")]
        cfg.c_files.extend(
            [str(file) for file in Path(self._code_gen.generated_code_dir).glob("*.c")]
        )
        cfg.include_dirs = [str(self._target_dir()), str(self._code_gen.generated_code_dir)]
        cfg.o_files = [str(file) for file in self._target_dir().glob("*.def")]
        cfg.targets = [Target(self._out_name, TargetKind.SHARED_LIBRARY)]
        builder = BuildSystem(self._project.app.install_dir)
        result = builder.build(cfg)
        if not result.success:
            raise ScadeOneException(f"Error building wrapper: {result.messages}")
        shutil.copy(build_dir / f"{self._out_name}.dll", self._target_dir())
        shutil.rmtree(build_dir)

    def generate(self) -> None:
        """
        Generate wrapper for Scade One project.
        """

        # Initialize data
        _files = []
        self._init_data()
        # Clean target dir before generating new files
        shutil.rmtree(self._target_dir())
        # Generate swan_cg_files.txt for files generated by KCG swan
        _swan_cg_file = self._target_dir() / "swan_cg_files.txt"
        _start = _swan_cg_file.parent.absolute()
        with _swan_cg_file.open("w") as f:
            # add files generated by Swan Code Generator
            for file_name in self._generate_declaration_files():
                f.write(
                    (self._target_dir() / file_name)
                    .absolute()
                    .relative_to(_start)
                    .as_posix()
                    .replace("\\", "/")
                    + "\n"
                )
        # Generate *.def file
        _def_pathname = self._target_dir() / (f"{self._out_name}.def")
        _files.append(_def_pathname)
        self._generate_def(_def_pathname)
        # Generate *.c file
        _c_pathname = self._target_dir() / (f"{self._out_name}.c")
        _files.append(_c_pathname)
        self._generate_c_code(_c_pathname)
        # Generate *.py file
        _py_pathname = self._target_dir() / (f"{self._out_name}.py")
        _files.append(_py_pathname)
        self._generate_python(_py_pathname)
        # Generate swan_config.h file
        _conf_pathname = self._target_dir() / "swan_config.h"
        _files.append(_conf_pathname)
        self._generate_swan_config(_conf_pathname)
        # Generate py_wrapper_files.txt
        _py_wrapper_files = self._target_dir() / "py_wrapper_files.txt"
        _start_file = _py_wrapper_files.parent.absolute()
        _cnt = "\n".join(f.absolute().relative_to(_start_file).as_posix() for f in _files)
        _py_wrapper_files.write_text(_cnt)
        # Build wrapper (DLL)
        self._build_wrapper()



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import OrderedDict
import functools
from pathlib import Path
from typing import Optional, Union, Tuple, cast
import re
from ansys.scadeone.core import ScadeOne
from ansys.scadeone.core.model.model import Model
import ansys.scadeone.core.svc.simdata as sd
import ansys.scadeone.core.swan as swan

from .utils import ConverterLogger

from .sss_parser import SSSParser


class SwanContext:
    """Class for Swan context"""

    RenamingRe = re.compile(r"(?P<kind>\w+)\s+(?P<old>\S+)\s+is\s+now\s+(?P<new>\S+)")

    def __init__(
        self,
        project_path: str,
        s_one_path: Union[str, None],
        renamings: Union[str, None],
    ) -> None:
        self._s_one_install = s_one_path
        self._project_path = project_path
        self._renamings_path = renamings
        self._renaming_dict = {}
        self._operator: swan.Operator = None
        self._model: Model = None
        self._app: ScadeOne = None
        self._inputs = OrderedDict()
        self._outputs = OrderedDict()
        self._module: swan.Module = None

    @property
    def s_one_install(self):
        return self._s_one_install

    @property
    def inputs(self) -> OrderedDict:
        return self._inputs

    @property
    def outputs(self) -> OrderedDict:
        return self._outputs

    @property
    def operator_module(self):
        # path of the module of the current root operator
        return self._module.get_full_path()

    @property
    def operator_path(self):
        return self._operator.get_full_path()

    @property
    def project_basename(self) -> str:
        return Path(self._project_path).stem

    def _load_renamings(self, renamings_path: str):
        with open(renamings_path, "r") as f:
            for line in f:
                m = SwanContext.RenamingRe.match(line.strip())
                if m is None:
                    continue
                kind = m.group("kind")
                if kind in ("function", "node"):
                    kind = "operator"
                if kind not in self._renaming_dict:
                    self._renaming_dict[kind] = {}
                self._renaming_dict[kind][m.group("old")] = m.group("new")

    def get_renamed(self, kind: str, name: str) -> Tuple[str, bool]:
        if not (self._renaming_dict) and self._renamings_path:
            self._load_renamings(self._renamings_path)
        try:
            return (self._renaming_dict[kind][name], True)
        except KeyError:
            # not found, add the project name if needed
            if name.find("::") == -1:
                name = self.project_basename + "::" + name
            return (name, False)

    def get_operator(self, op_name: str):
        ConverterLogger.getLogger("Loading operator")
        self._app = ScadeOne(install_dir=self.s_one_install) if self._s_one_install else ScadeOne()
        project = self._app.load_project(self._project_path)
        if not project:
            ConverterLogger.exception("Invalid project")
        self._model = self._app.model
        (full_name, is_renamed) = self.get_renamed("operator", op_name)
        module_part, name = swan.PathIdentifier.split(full_name)
        module = self._model.get_module_body(module_part)
        if module is None:
            ConverterLogger.exception(f"Cannot find module {module_part}")
        self._module = module
        operator = module.get_declaration(name)
        if operator is None or not isinstance(operator, swan.Operator):
            ConverterLogger.exception(f"Cannot find operator {op_name} (stp 'operator' attribute)")
        self._operator = operator

        for sig in operator.inputs:
            if isinstance(sig.type, swan.TypeGroupTypeExpression):
                sig_name = sig.id.value
                full_path_name = f"{operator.get_full_path()}/{sig_name}"
                type = self.get_sd_type_of_type_expr(sig.type.type, module)
                self._inputs[full_path_name] = type

        for sig in operator.outputs:
            if isinstance(sig.type, swan.TypeGroupTypeExpression):
                sig_name = sig.id.value
                full_path_name = f"{operator.get_full_path()}/{sig_name}"
                type = self.get_sd_type_of_type_expr(sig.type.type, module)
                self._outputs[full_path_name] = type

    def get_sensor_sd_type(self, sensor_path: str) -> Union[Tuple[str, sd.Type], None]:
        """Looks for a sensor in its module or in the project module.

        Returns its type , or None if not found."""

        for sensor in self._model.sensors:
            if sensor.get_full_path() == sensor_path:
                return self.get_sd_type_of_type_expr(sensor.type, self._module)
        return None

    def get_sd_type_of_type_expr(
        self,
        type_expr: swan.TypeExpression,
        module: swan.Module,
        name: Optional[str] = None,
    ):
        if isinstance(type_expr, swan.ArrayTypeExpression):
            res = self.get_array_type(type_expr, module)
            if res is None:
                return None
            sd_type, dims = res
            sd_array_type = sd.create_array_type(sd_type, dims, name if name else "")
            return sd_array_type

        if isinstance(type_expr, swan.PredefinedType):
            # bool & cie
            sd_type = self.get_sd_predefined_type(type_expr)

            return sd_type

        if isinstance(type_expr, swan.TypeReferenceExpression):
            # type name: T1, M1::T
            type_name = type_expr.alias.as_string
            type_decl = module.get_declaration(type_name)
            return self.get_sd_type_of_type_decl(type_decl, type_decl.module)

        if isinstance(type_expr, swan.SizedTypeExpression):
            # signed<<expr>> or unsigned <<expr>>
            ConverterLogger.warning("Unsupported Sized Type was found for type expression.")
            return None

        if isinstance(type_expr, swan.VariableTypeExpression):
            # 'T: should never occur at top-level interface
            ConverterLogger.warning("Unsupported Variable Type was found for type expression.")
            return None

        ConverterLogger.error("Undefined Type was found for type expression.")
        return None

    @functools.cache
    def get_sd_type_of_type_decl(self, decl: swan.TypeDecl, module: swan.Module):
        name = decl.get_full_path()
        type_def = decl.definition

        if isinstance(type_def, swan.ExprTypeDefinition):
            sd_type = self.get_sd_type_of_type_expr(type_def.type, module, name)

        elif isinstance(type_def, swan.EnumTypeDefinition):
            enum_module, enum_type = swan.PathIdentifier.split(name)
            tags = [enum_module + "::" + i.value for i in type_def.tags]
            sd_type = sd.create_enum_type(tags, name)

        elif isinstance(type_def, swan.StructTypeDefinition):
            sd_fields = []
            for f in type_def.fields:
                field_value = self.get_sd_type_of_type_expr(f.type, module)
                if field_value is None:
                    return None
                field_name = f.id.value
                sd_fields.append((field_name, field_value))
            sd_type = sd.create_struct_type(sd_fields, name)

        else:
            ConverterLogger.error(f"Unsupported imported type: {name}")
            return None

        return sd_type

    def get_array_type(self, array_type: swan.ArrayTypeExpression, module: swan.Module):
        size = array_type.size
        if isinstance(size, swan.Literal) and size.is_integer:
            size = int(size.value)
        elif isinstance(size, swan.PathIdExpr):
            path = size.path_id.as_string
            size = self.get_constant_value(path, module)
            if not (isinstance(size, int) and size >= 0):
                ConverterLogger(f"Constant {path} value is not an integer")
                return None
        else:
            ConverterLogger.error(f"Cannot handle size: {size}")
            return None
        type = array_type.type
        if isinstance(type, swan.ArrayTypeExpression):
            res = self.get_array_type(type, module)
            if res is None:
                return None
            (sd_type, dims) = res
            dims.insert(0, size)
        else:
            sd_type = self.get_sd_type_of_type_expr(type, module)
            dims = [size]
        return (sd_type, dims)

    def get_sd_predefined_type(self, predefined):
        if type(predefined) is swan.BoolType:
            return sd.Bool
        if type(predefined) is swan.Int8Type:
            return sd.Int8
        if type(predefined) is swan.Int16Type:
            return sd.Int16
        if type(predefined) is swan.Int32Type:
            return sd.Int32
        if type(predefined) is swan.Int64Type:
            return sd.Int64
        if type(predefined) is swan.Uint8Type:
            return sd.UInt8
        if type(predefined) is swan.Uint16Type:
            return sd.UInt16
        if type(predefined) is swan.Uint32Type:
            return sd.UInt32
        if type(predefined) is swan.Uint64Type:
            return sd.UInt64
        if type(predefined) is swan.Float32Type:
            return sd.Float32
        if type(predefined) is swan.Float64Type:
            return sd.Float64
        if type(predefined) is swan.CharType:
            return sd.Char
        ConverterLogger.error(f"Unknown Predefined Type: {predefined}")
        return None

    def module_look_for(self, module_name, search_err_msg, search_fn):
        """Apply search_fn to find"""
        body = self._model.get_module_body(module_name)
        interface = self._model.get_module_interface(module_name)
        if not (body or interface):
            ConverterLogger.error(f"Module {module_name} not found for {search_err_msg}")
            return None
        if body:
            item = search_fn(body)
            if item:
                return item
        if interface:
            item = search_fn(interface)
            if item:
                return item
        return None

    def get_enum(self, name: str) -> Union[str, None]:
        """Look for an enum without module part"""
        full_name = name if name.find("::") != -1 else self.project_basename + "::" + name
        module_part, enum_tag = swan.PathIdentifier.split(full_name)
        # Look for renaming of the enum. Note this is a type renaming, so we need to look for
        # the module part in the type name
        for t in self._renaming_dict.get("type", {}).keys():
            (type_module, _) = swan.PathIdentifier.split(t)
            if type_module == module_part:
                (new_module, _) = swan.PathIdentifier.split(self._renaming_dict["type"][t])
                module_part = new_module
                break

        def search_fn(m: swan.Module):
            for typ in m.types:
                if not isinstance(typ.definition, swan.EnumTypeDefinition):
                    continue
                for tag in typ.definition.tags:
                    if tag.value == enum_tag:
                        return full_name
            return None

        enum = self.module_look_for(module_part, f"enum {name}", search_fn)

        return enum

    def _eval_constant_expr(self, cst_value: swan.Expression):
        if isinstance(cst_value, swan.Literal):
            if cst_value.is_bool:
                return cst_value.is_true
            elif cst_value.is_char:
                return SSSParser.to_char(cst_value.value)
            elif cst_value.is_integer:
                return int(cst_value.value)
            elif cst_value.is_float:
                return float(cst_value.value)
            else:
                None
        elif isinstance(cst_value, swan.UnaryExpr):
            unary = cast(swan.UnaryExpr, cst_value)
            value = self._eval_constant_expr(unary.expr)
            if not isinstance(value, (int, float)):
                return None
            return -value if unary.operator == swan.UnaryOp.Minus else value

    def get_constant(
        self,
        name: str,
        module: Optional[swan.Module] = None,
        no_const_ok: Optional[bool] = False,
    ) -> Union[swan.Literal, None]:
        """Look for an enum without module part"""
        const_decl = None

        if module is None:
            # case of a constant from .sss
            (full_name, is_renamed) = self.get_renamed("constant", name)

            module_part, cst_name = swan.PathIdentifier.split(full_name)

            def search_fn(m: swan.Module):
                const_decl = None
                for const in m.constants:
                    if const.id.value != cst_name:
                        continue
                    const_decl = const
                    break
                return const_decl

            const_decl = self.module_look_for(module_part, f"constant {name}", search_fn)

        else:
            # get constant from given module
            const_decl = module.get_declaration(name)

        if const_decl is None and not no_const_ok:
            ConverterLogger.error(f"Constant {name} not found")

        return const_decl

    def get_constant_value(
        self,
        name: str,
        module: Optional[swan.Module] = None,
        no_const_ok: Optional[bool] = False,
    ) -> Union[swan.Literal, None]:
        """Look for an enum without module part"""
        const_decl = self.get_constant(name, module, no_const_ok)
        if const_decl is None:
            return None
        if const_decl.value is None:
            ConverterLogger.error(f"Imported constant {name} is not supported")
            return None
        value = self._eval_constant_expr(const_decl.value)
        if value is None:
            ConverterLogger.error(
                f"Cannot evaluate constant: {name} (must be a numeric, bool or char value)"
            )
            return None
        return value



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import ctypes

from enum import IntEnum


###########################################################################
# .sd types & constants
sd_bool_t = ctypes.c_ubyte
sd_bool_true = ctypes.c_ubyte(1)
sd_bool_false = ctypes.c_ubyte(0)
sd_byte_t = ctypes.c_ubyte

sd_int8_t = ctypes.c_int8
sd_uint8_t = ctypes.c_uint8
sd_int16_t = ctypes.c_int16
sd_uint16_t = ctypes.c_uint16
sd_int32_t = ctypes.c_int32
sd_uint32_t = ctypes.c_uint32
sd_int64_t = ctypes.c_int64
sd_uint64_t = ctypes.c_uint64
sd_float32_t = ctypes.c_float
sd_float64_t = ctypes.c_double

sd_id_t = ctypes.c_int64
sd_err_t = ctypes.c_int32
sd_size_t = ctypes.c_size_t
sde_kind_t = ctypes.c_int32
sdt_class_t = ctypes.c_int32
sdd_value_class_t = ctypes.c_int32
sdd_sequence_class_t = ctypes.c_int32
sd_value_iterator_t = ctypes.c_void_p
sd_value_t = ctypes.c_void_p
sd_sequence_t = ctypes.c_void_p
sdf_open_mode_t = ctypes.c_int32

sd_pfn_vsize_get_bytes_size_t = ctypes.CFUNCTYPE(sd_size_t, ctypes.c_void_p)
sd_pfn_vsize_to_bytes_t = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(sd_byte_t))

SD_ID_INVALID = -1
SD_SIZE_NONE = 2**64 - 1
SD_TAG_VALUE_NONE = -1
SD_ERR_NONE = 0
SDT_NONE = SD_ID_INVALID


class TypeClass(IntEnum):
    UNDEFINED = SD_ID_INVALID
    PREDEF = 0
    STRUCT = 1
    ARRAY = 2
    ENUM = 3
    VARIANT = 4
    IMPORTED = 5


class PredefinedType(IntEnum):
    CHAR = 0
    # CHAR16 = 1 # Reserved
    # CHAR32 = 2 # Reserved
    BOOL = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    INT64 = 7
    UINT8 = 8
    UINT16 = 9
    UINT32 = 10
    UINT64 = 11
    FLOAT32 = 12
    FLOAT64 = 13


class SdeKind(IntEnum):
    NONE = SD_ID_INVALID
    SENSOR = 0
    INPUT = 1
    OUTPUT = 2
    WIRE = 3
    LOCAL = 4
    PROBE = 5
    GROUP_ITEM = 6
    ASSUME = 7
    GUARANTEE = 8
    SIGNAL = 9
    OPERATOR = 10
    INSTANCE = 11
    AUTOMATON = 12
    STATE = 13
    FORK = 14
    TRANSITION = 15
    IF_BLOCK = 16
    WHEN_BLOCK = 17
    BRANCH = 18
    ASSERT = 19


class DataClass(IntEnum):
    UNDEFINED = -1
    NONE = 0
    PREDEF = 1
    LIST = 2
    ENUM = 3
    VARIANT = 4
    UNTYPED_VARIANT_CONSTRUCTOR = 5
    IMPORTED = 6


class FileOpenMode(IntEnum):
    READ = 0
    CREATE = 1
    EDIT = 2



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC
from typing import TYPE_CHECKING, cast

from ansys.scadeone.core.common.storage import ProjectStorage
from ansys.scadeone.core.interfaces import IScadeOne
from ansys.scadeone.core.common.exception import ScadeOneException

if TYPE_CHECKING:
    from ansys.scadeone.core.project import Project
    from ansys.scadeone.core.scadeone import ScadeOne
    from ansys.scadeone.core.swan.modules import Module, ModuleInterface


class ProjectFactory:
    @staticmethod
    def create_project(app: "IScadeOne", storage: ProjectStorage) -> "Project":
        """Create a project."""
        from ansys.scadeone.core.project import Project

        project = Project(app, storage)
        project.is_new = True
        return project


class ProjectAdder:
    @staticmethod
    def add_project(app: "ScadeOne", project: "Project") -> None:
        """Add a project to the application."""
        app.projects.append(project)


class ProjectCreator(ABC):
    @staticmethod
    def _set_module_source(project: "Project", module: "Module") -> None:
        project_dir = project.directory
        if not project_dir:
            raise ScadeOneException("Cannot add module to project without directory.")
        swan_file_name = module.file_name
        module_path = project_dir / "assets" / swan_file_name
        module.source = str(module_path)

    def add_module(self, name: str) -> "Module":
        """Add a module to the project.

        Parameters
        ----------
        name: str
            Module name.

        Returns
        -------
        Module
            Module object.
        """
        from ansys.scadeone.core.svc.swan_creator.module_creator import ModuleAdder, ModuleFactory

        module = ModuleFactory.create_module(name)
        ModuleAdder.add_module(self.model, module)
        ProjectCreator._set_module_source(cast("Project", self), module)
        return module

    def add_module_interface(self, name: str) -> "ModuleInterface":
        """Add a module interface to the model.

        Parameters
        ----------
        name: str
            Module name.

        Returns
        -------
        ModuleInterface
            ModuleInterface object.
        """
        from ansys.scadeone.core.svc.swan_creator.module_creator import ModuleAdder, ModuleFactory

        module = ModuleFactory.create_module_interface(name)
        ModuleAdder.add_module(self.model, module)
        ProjectCreator._set_module_source(cast("Project", self), module)
        return module



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains classes for package and interface.
"""

from typing import List, Optional, Union, cast
from pathlib import Path

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.svc.swan_creator.module_creator import ModuleCreator
import ansys.scadeone.core.swan.common as common


from .globals import ConstDecl, SensorDecl
from .groupdecl import GroupDecl
from .operators import Operator, Signature
from .typedecl import TypeDecl


class GlobalDeclaration(common.ModuleItem):  # numpydoc ignore=PR01
    """Abstract class for global declarations:

    - type declaration list
    - constant declaration list
    - sensor declaration list
    - group declarations
    - user operator declaration (without body, in interface)
    - user operator definition (with body)
    """

    def __init__(self) -> None:
        super().__init__()

    def get_full_path(self) -> str:
        """Full path of Swan construct."""
        if self.owner is None:
            raise ScadeOneException("No owner")
        return f"{self.owner.get_full_path()}"


class TypeDeclarations(GlobalDeclaration):  # numpydoc ignore=PR01
    """Type declarations: **type** {{ *type_decl* ; }}."""

    def __init__(self, types: List[TypeDecl]) -> None:
        super().__init__()
        self._types = types
        common.SwanItem.set_owner(self, types)

    @property
    def types(self) -> List[TypeDecl]:
        """Declared types."""
        return self._types


class ConstDeclarations(GlobalDeclaration):  # numpydoc ignore=PR01
    """Constant declarations: **constant** {{ *constant_decl* ; }}."""

    def __init__(self, constants: List[ConstDecl]) -> None:
        super().__init__()
        self._constants = constants
        common.SwanItem.set_owner(self, constants)

    @property
    def constants(self) -> List[ConstDecl]:
        """Declared constants."""
        return self._constants


class SensorDeclarations(GlobalDeclaration):  # numpydoc ignore=PR01
    """Sensor declarations: **sensor** {{ *sensor_decl* ; }}."""

    def __init__(self, sensors: List[SensorDecl]) -> None:
        super().__init__()
        self._sensors = sensors
        common.SwanItem.set_owner(self, sensors)

    @property
    def sensors(self) -> List[SensorDecl]:
        """Declared sensors."""
        return self._sensors


class GroupDeclarations(GlobalDeclaration):  # numpydoc ignore=PR01
    """Group declarations: **group** {{ *group_decl* ; }}."""

    def __init__(self, groups: List[GroupDecl]) -> None:
        super().__init__()
        self._groups = groups
        common.SwanItem.set_owner(self, groups)

    @property
    def groups(self) -> List[GroupDecl]:
        """Declared groups."""
        return self._groups


class UseDirective(common.ModuleItem):  # numpydoc ignore=PR01
    """Class for **use** directive."""

    def __init__(
        self, path: common.PathIdentifier, alias: Optional[common.Identifier] = None
    ) -> None:
        super().__init__()
        self._path = path
        self._alias = alias

    @property
    def path(self) -> common.PathIdentifier:
        """Used module path."""
        return self._path

    @property
    def alias(self) -> Union[common.Identifier, None]:
        """Renaming of module."""
        return self._alias


class ProtectedDecl(common.ProtectedItem, GlobalDeclaration):  # numpydoc ignore=PR01
    """Protected declaration."""

    def __init__(self, markup: str, data: str) -> None:
        super().__init__(data, markup)

    @property
    def is_type(self) -> bool:
        """Protected type declaration."""
        return self.markup == "type"

    @property
    def is_const(self) -> bool:
        """Protected const declaration."""
        return self.markup == "const"

    @property
    def is_group(self) -> bool:
        """Protected group declaration."""
        return self.markup == "group"

    @property
    def is_sensor(self) -> bool:
        """Protected sensor declaration."""
        return self.markup == "sensor"

    @property
    def is_user_operator(self) -> bool:
        """Protected operator declaration.

        Note: operator declaration within {text% ... %text} is parsed."""
        return self.markup == "syntax_text"

    def get_full_path(self) -> str:
        """Full path of Swan construct."""
        if self.owner is None:
            raise ScadeOneException("No owner")
        return f"{self.owner.get_full_path()}::<protected>"


class Module(common.ModuleBase, ModuleCreator):  # numpydoc ignore=PR01
    """Module base class.

    Parameters
    ----------
    name : common.PathIdentifier
        module name
    use_directives : Union[List[UseDirective], None]
        **use** directives
    declarations : Union[List[common.ModuleItem], None]
        module declarations
    """

    def __init__(
        self,
        name: common.PathIdentifier,
        use_directives: List[UseDirective] = None,
        declarations: List[common.ModuleItem] = None,
    ) -> None:
        super().__init__()
        self._name = name
        self._uses = use_directives if use_directives else []
        self._declarations = declarations if declarations else []
        self._source = None
        common.SwanItem.set_owner(self, self._uses)
        common.SwanItem.set_owner(self, self._declarations)

    @property
    def name(self) -> common.PathIdentifier:
        """Module or Interface name."""
        return self._name

    @property
    def source(self) -> Union[str, None]:
        "Source of the module, as a string (file name)."
        return self._source

    @source.setter
    def source(self, path: str):
        "Set source of the module"
        self._source = path

    @property
    def declarations(self) -> List[common.ModuleItem]:
        """Module's declarations"""
        return self._declarations

    @property
    def use_directives(self) -> List[UseDirective]:
        """Module's **use** directives as a list."""
        return self._uses

    @property
    def extension(self) -> str:
        """Return module extension, with . included."""
        return ""

    @property
    def file_name(self) -> str:
        """Return a file name based on module name and namespaces."""
        return self.name.as_string.replace("::", "-") + self.extension

    @staticmethod
    def module_name_from_path(path: Path) -> str:
        """Returns a module name from a path.

        Parameters
        ----------
        path : Path
            Path to the module, e.g. */path/to/namespace-module.swan*

        Returns
        -------
        str
            Module name, e.g. *namespace::module*
        """
        module_name = path.stem.replace("-", "::")
        if common.PathIdentifier.is_valid_path(module_name):
            return module_name
        raise ScadeOneException(f"Invalid module path name: {path}")

    def filter_declarations(self, filter_fn) -> List[GlobalDeclaration]:
        """Return declarations matched by a filter.

        Parameters
        ----------
        filter_fn : function
            A function of one argument of type GlobalDeclaration, returning True or False.

        Returns
        -------
        List[GlobalDeclaration]
            List of matching declarations.
        """
        return list(filter(filter_fn, self.declarations))

    @property
    def types(self) -> List[TypeDecl]:
        """Return a list of type declarations."""
        types = []
        for decl in self.filter_declarations(lambda x: isinstance(x, TypeDeclarations)):
            for typ in cast(TypeDeclarations, decl).types:
                types.append(typ)
        return types

    @property
    def sensors(self) -> List[SensorDecl]:
        """Return a list of sensor declarations."""
        sensors = []
        for decl in self.filter_declarations(lambda x: isinstance(x, SensorDeclarations)):
            for sensor in cast(SensorDeclarations, decl).sensors:
                sensors.append(sensor)
        return sensors

    @property
    def constants(self) -> List[ConstDecl]:
        """Return a list of constant declarations."""
        constants = []
        for decl in self.filter_declarations(lambda x: isinstance(x, ConstDeclarations)):
            for constant in cast(ConstDeclarations, decl).constants:
                constants.append(constant)
        return constants

    @property
    def groups(self) -> List[GroupDecl]:
        """Return a list of group declarations."""
        groups = []
        for decl in self.filter_declarations(lambda x: isinstance(x, GroupDeclarations)):
            for grp in cast(GroupDeclarations, decl).groups:
                groups.append(grp)
        return groups

    @property
    def signatures(self) -> List[Signature]:
        """Return a list of signatures."""
        return [
            cast(Signature, signature)
            for signature in self.filter_declarations(lambda x: isinstance(x, Signature))
            if signature
        ]

    def get_full_path(self) -> str:
        """Full Swan path of module."""
        return self.name.as_string

    def get_declaration(self, name: str) -> GlobalDeclaration:
        """Return the type, global, or operator declaration searching by namespace."""
        from .namespace import ModuleNamespace

        m_ns = ModuleNamespace(self)
        decl = m_ns.get_declaration(name)
        return cast(GlobalDeclaration, decl) if decl else None

    def get_use_directive(self, module_name: str) -> Union[UseDirective, None]:
        """Return the use directive searching by module name or alias name.

        Aliases are searched first, the module name is searched if no alias is found.
        When a module name is searched, use directive with first occurrence is returned.

        Parameters
        ----------
            module_name : str
                Name of the module (last part of a path) or alias.

        Returns
        -------
            Found UseDirective or None
        """
        for use in self.use_directives:
            if use.alias:
                key = use.alias.value
            else:
                if not isinstance(use.path.path_id, list):
                    raise ScadeOneException(f"{use.path.as_string} is invalid.")
                key = use.path.path_id[-1].value
            if key == module_name:
                return use
        return None

    def interface(self) -> "ModuleInterface":
        """Return the module interface for a module body if it exists."""
        return None

    def body(self) -> "ModuleBody":
        """Return the module body for a module interface if it exists."""
        return None

    def harness(self):
        """Return the test harness if it exists."""
        return None


class ModuleInterface(Module):  # numpydoc ignore=PR01
    """Module interface definition."""

    def __init__(
        self,
        name: common.PathIdentifier,
        use_directives: Optional[List[UseDirective]] = None,
        declarations: Optional[List[common.ModuleItem]] = None,
    ) -> None:
        super().__init__(name, use_directives, declarations)

    @property
    def extension(self) -> str:
        """Return module extension, with . included."""
        return ".swani"

    def body(self) -> "ModuleBody":
        """Return the module body for a module interface if it exists."""
        return self.model.get_module_body(self.name.as_string)


class ModuleBody(Module):  # numpydoc ignore=PR01
    """Module body definition."""

    def __init__(
        self,
        name: common.PathIdentifier,
        use_directives: Optional[List[UseDirective]] = None,
        declarations: Optional[List[GlobalDeclaration]] = None,
    ) -> None:
        super().__init__(name, use_directives, declarations)

        from ansys.scadeone.core.model.loader import SwanParser

        self._parser = SwanParser(LOGGER)

    @property
    def extension(self) -> str:
        """Return module extension, with '.' included."""
        return ".swan"

    @property
    def operators(self) -> List[Operator]:
        """Return a list of operators."""
        return [
            cast(Operator, decl)
            for decl in self.filter_declarations(lambda x: isinstance(x, Operator))
            if decl
        ]

    def interface(self) -> "ModuleInterface":
        """Return the module interface for a module body if it exists."""
        return self.model.get_module_interface(self.name.as_string)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cspell: ignore startuml enduml unmark
# pylint: disable=W0621

"""
This module defines a *Document* concept and render a document.
A Document is composed of DElt instances which consist in:

- printable information (text, newline)
- indentation information
- group or list of DElt
- and the next DElt.

Therefore DElt items form a tree structure:
- sequence of DElt, thanks to the next DElt contained in each DElt
- Structure with the DBlock, which contains DElt.

For instance::

    (
        Hello World !
        Hi there
    )

can be conceptually represented as:

.. uml::

    @startuml
    hide circle
    hide method

    class LPar <<DText>> {
    {field} text: "("
    next
    }

    class indent <<DIndent>> {
    next
    }

    class nl <<DLineBreak>> {
    next
    }

    class list <<DList>> {
    next
    }

    class b1 <<DBlock>> {
    doc
    }

    class hello <<DText>> {
    text: "Hello "
    next
    }
    class world<<DText>> {
    text: "World !"
    }

    class b2 <<DBlock>> {
    doc
    }

    class hi<<DText>> {
    text: "Hi"
    next
    }

    class there <<DText>> {
    text: "there"
    }


    class unindent <<DIndent>> {
    next
    }

    class nl2 <<DLineBreak>> {
    next
    }

    class RPar <<DText>> {
    {field} text: ")"
    }

    hello::next --> world
    b1::doc --> hello
    hi::next --> there
    b2::doc --> hi
    list --> b1
    list --> b2
    together {
       class b1
       class b2
    }

    LPar::next --> indent
    indent::next --> nl
    nl::next --> list
    list::next --> unindent
    unindent --> nl2
    nl2 --> RPar

    @enduml


Methods
-------

The :py:class:`Document` defines the main object. A
document is rendered by a :py:class:`Render`object.

Once a document object is created, several operations are
provided to build the document. Operations are provide as
functions.

String shortcuts can be used as kind of interpreted commands.
A shortcut is prefixed with a **@** character and may
have options separated with **:**

Methods, with their shortcut after the slash are:

text(str)
    Create a text doc element.

indent(), unindent() / i, u
    Add or remove an indentation level. The level of the
    indentation is handle by the render object.

mark(), unmark() / m, M
    Set/unset the next indentation to the column of the last
    rendered document. This can be used to align lines after
    a parenthesis

nl(bool = True) / n
    Insert a line break. It has a boolean option which inserts
    an indentation right after a line break.

    Shortcut is "nl:t" or "nl:f" for True or False

sep(str, bool = False) / s
    Create a new document to be used as a separator for lists.
    The optional boolean inserts a newline with indentation if
    True.

    Shortcut is "@s:<txt>" where txt is the separator. For instance
    "@s:;" defines ";" as a separator. A line break can be added
    line the *nl* command with the syntax "@s:;:t"

block(doc, ...)
    Create a block with the docs given in arguments.
    A block contains the docs, and is itself a doc element.
    It is used to create structure in document.
    A block itself is a document element that can be
    connected to other elements.

doc_list(doc, ..., sep=, start=, last=)
    Create a list of document elements, with a possible
    element separator and start, end separators element.

    Newline can be handled by the separator element.
    A list is itself a document element that can be
    connected to other elements


Note: as *nl* and *sep* commands may generate an indentation,
the *indent* must be called before, as it "pushes" the new
indentation.

Document element can be added to the main document using '<<'.
The '<<' operator adds the document element and returns the
main document. The argument can be a string, a shortcut, or
an other document element:
Example:

>>> doc = Document()
>>> x = doc << "Hello" << " " << "world"
>>> x == doc
True

A document element has also a '<<' operator which set
the argument as next element and return it (different from
*Document.<<()*). Arguments are DElt only

>>> h = text("Hello")
>>> w = text("World")
>>> d = h << w
>>> d == w
True

Document rendering:

>>> import sys
>>> r = Renderer(sys.stdout)
>>> r.render(doc)
Hello world

Examples
--------

Creation of a document. We use "@m" and "@M" to mark
specific indentation after '('. The "@s" separator is
is used with and without newline:

>>> def build_doc():
...   doc = Document()
...   doc << "node"
...   doc << " " << "Op" << "(" << "@m"
...   doc << doc_list("1", "2", "3", sep="@s:;:t", start="/* in */") << ")"
...   doc << "@M" << "@i" << "@n" << "return (" << "@m"
...   io = [
...      block("a", ": ", "T", ";"),
...      block("b", ": ", "T", ";"),
...   ]
...   doc << doc_list(*io, sep="@n") << ")" << "@M" << "@u" << "@n"
...   doc << "{" << "@i" << "@n"
...   doc << "let" << "@m" << "@n"
...   doc << "/* hello */" << "@n"
...   doc << "/* world */" << "@M"
...   doc << "@u" << "@n" << "}" << "@n"
...   return doc

Rendering:
>>> Renderer(sys.stdout).render(build_doc())
node Op(/* in */1;
        2;
        3)
  return (a: T;
          b: T;)
{
  let
     /* hello */
     /* world */
}

The :py:class:`Renderer` uses a stream as parameter,
which can be a stream buffer, of file, ... It can
be also derived to implement other kind of rendering.
"""  # numpydoc ignore

from enum import Enum, auto
import re
from typing import Any, List, Optional, Union, IO, Callable

from ansys.scadeone.core.common.exception import ScadeOneException


class DElt:
    """Document base class."""

    def __init__(self) -> None:
        self._next: Optional["DElt"] = None

    def cons(self, doc: "DElt") -> "DElt":
        """Add next element and return it."""
        self._next = doc
        return doc

    @property
    def next(self) -> Union["DElt", None]:
        """Return the next element."""
        if self._next == self:
            raise ScadeOneException(f"Document connected to itself: {repr(self)}")
        return self._next

    def __lshift__(self, doc) -> "DElt":
        """<< operator: Add doc as next element and return it."""
        doc = to_doc(doc)
        return self.cons(doc)


class DText(DElt):
    """Class to store a simple text."""

    def __init__(self, string: str) -> None:
        super().__init__()
        if not isinstance(string, str):
            raise ScadeOneException("DText(): string expected")
        self._string = string

    @property
    def string(self) -> str:
        """Return the text string."""
        return self._string

    def __repr__(self) -> str:
        return super().__repr__() + f'("{self._string}")'


class DBlock(DElt):
    """Class to store a block of documents (sequence of documents)."""

    def __init__(self, doc: Optional[DElt] = None) -> None:
        super().__init__()
        self._first = doc if doc else None
        self._end = self._first

    @property
    def doc(self) -> DElt:
        """Return the first document of the block."""
        return self._first

    def __lshift__(self, doc) -> "DBlock":
        """<< operator: Add doc to the end of the block and return the block."""
        doc = to_doc(doc)
        if not self._end:
            self._first = doc
            self._end = doc
        else:
            self._end = self._end.cons(doc)
        return self


class DLineBreak(DElt):
    """Class to store a line break."""

    def __init__(self, with_indent=True) -> None:
        super().__init__()
        self._with_indent = with_indent

    @property
    def with_indent(self) -> bool:
        """Return True if indentation is added after the line break."""
        return self._with_indent

    def __repr__(self) -> str:
        return super().__repr__() + f"({self._with_indent})"


class EIndentation(Enum):
    """Kind of indentation."""

    #: No indentation
    NOTHING = auto()
    #: Start new indentation
    INDENT = auto()
    #: End last indentation, including mark
    UNINDENT = auto()
    #: Mark last column for indentation
    MARK = auto()


class DIndent(DElt):
    """Class to store an indentation."""

    def __init__(self, indent: EIndentation) -> None:
        super().__init__()
        self._indent = indent

    @property
    def indent(self) -> EIndentation:
        """Return the kind of indentation."""
        return self._indent

    def __repr__(self) -> str:
        return super().__repr__() + f"({self._indent})"


# Document creation functions
CmdRe = re.compile("^(?P<c>@[a-zA-Z]+)(?P<a>(?::[^:])*)")
Cmd = {
    "@i": "indent",
    "@u": "unindent",
    "@m": "mark",
    "@M": "unmark",
    "@n": "nl",
    "@s": "sep",
}


def fix_arg(arg):
    """Transform arg into bool, int or keep it"""
    if arg == "t":
        return True
    if arg == "f":
        return False
    if isinstance(arg, int):
        return int(arg)
    return arg


def cmd(arg: str) -> Union[DElt, None]:
    """Check if arg is a document manipulation string command
    and returns its result if it is the case, else None."""
    if not isinstance(arg, str):
        return None
    m = CmdRe.match(arg)
    if not m:
        return None
    if m["c"] not in Cmd:
        return None
    func = Cmd[m["c"]]
    # first item will always be empty.
    args = m["a"].split(":")[1:]
    args = [fix_arg(arg) for arg in args]
    return func(*args)


def to_doc(arg: Union[DElt, str]) -> DElt:
    """Create a DElt:
    - If arg is already a DElt, return it
    - If arg is a shortcut, returns command
    - Else return a text()."""
    if isinstance(arg, DElt):
        return arg
    res = cmd(arg)
    if res:
        return res
    return text(arg if arg else "")


def text(txt: str, with_indent: bool = True) -> DElt:
    """Create a DElt from txt.
    If text contains newlines, they are replaced by line breaks,
    with indentation if with_indent is True."""
    if txt.find("\n") != -1:
        doc = DBlock()
        for line in txt.split("\n"):
            doc << text(line) << nl(with_indent)
        return doc
    return DText(txt)


def nl(with_indent: bool = True):
    """Add a new line. If `with_indent is True,
    indentation is added beginning of next line."""
    return DLineBreak(with_indent)


def indent() -> DElt:
    """Add an indentation. Indentation is set by Render."""
    return DIndent(EIndentation.INDENT)


def unindent() -> DElt:
    """Back to previous indentation. Indentation is set by Render."""
    return DIndent(EIndentation.UNINDENT)


def mark() -> DElt:
    """Mark last column as reference for indentation.

    Indented new lines use column-spaces. Column depends on
    latest rendered document."""
    return DIndent(EIndentation.MARK)


def unmark() -> DElt:
    """Remove previous mark"""
    return DIndent(EIndentation.UNINDENT)


def block(first: DElt, *args) -> DElt:
    """Create a block of n-documents. A block is assimilated
    as a single document. Each internal document are rendered in sequence.
    """
    first = to_doc(first)
    current_doc = first
    for doc in args:
        doc = to_doc(doc)
        current_doc.cons(doc)
        current_doc = doc
    return DBlock(first)


def doc_list(
    *docs: List[Union[DElt, str]],
    sep: Optional[Union[DElt, str]] = None,
    start: Optional[Union[DElt, str]] = None,
    last: Optional[Union[DElt, str]] = None,
) -> DElt:
    """Build a list of documents.

    Parameters
    ----------
        docs (List[Union[DElt, str]]): any number of documents
        sep (Optional[Union[DElt, str]], optional): separator. Defaults to None.
        start (Optional[Union[DElt, str]], optional): document at beginning of list.
        Defaults to None.
        last (Optional[AUnion[DElt, str]ny], optional): document at end of list.
        Defaults to None.

    Returns
    -------
        Document: _description_
    """
    if not docs:
        return DBlock()
    _init, *_rest = docs
    _current_doc = to_doc(_init)
    if start:
        _first = to_doc(start)
        _first.cons(_current_doc)
    else:
        _first = _current_doc
    _sep = to_doc(sep)
    for _arg in _rest:
        _doc = to_doc(_arg)
        if sep:
            _current_doc = _current_doc.cons(DBlock(_sep))
        _current_doc = _current_doc.cons(_doc)
    if last:
        _current_doc.cons(to_doc(last))
    return DBlock(_first)


def sep(char: str, newline: bool = False) -> DElt:
    """Helper to create a document separator (comma, semi-col, ...)."""
    doc = text(char)
    if not newline:
        return doc
    doc.cons(nl(True))
    return DBlock(doc)


# Fix function references
for k, v in Cmd.items():
    Cmd[k] = globals()[v]


class Document(DBlock):
    """Document object.

    A document object contains the starting DElt and has the '<<' method.

    *a_doc* << *d_elt* => add *d_elt* and returns *a_doc*."""

    def __init__(self) -> None:
        super().__init__()

    @property
    def start_doc(self):
        """Return the first document."""
        return self.doc

    @property
    def end_doc(self):
        """Return the last document."""
        return self._end


class Renderer:
    """Rendering class.

    Render a document on a stream. The stream can be a file, a buffer, ...
    The rendering is done by calling the render method which applies to
    a specific DElt.

    """

    def __init__(
        self,
        stream: Optional[IO] = None,
        indent: Optional[int] = 2,
        initial_indent: Optional[int] = 0,
    ) -> None:
        """
        Parameters
        ----------
            stream (io.IO): Output stream
            indent (int, optional): default indentation. Defaults to 2.
            initial_indent (int, optional): initial indentation. Defaults to 0.
        """
        self._indent = indent
        self._initial_indent = initial_indent
        self._col = 0
        self._line = 0
        self._indent_stack = [initial_indent]
        self._stream = stream

    def set_stream(self, stream: IO):
        """Set the stream to render on."""
        self._stream = stream

    @property
    def lines(self) -> int:
        """Current line number"""
        return self._line

    @property
    def default_indent(self):
        """Default indent, as given in constructor."""
        return self._indent

    @property
    def curr_indent(self):
        """Current indent"""
        if self._indent_stack:
            return self._indent_stack[-1]
        raise ScadeOneException("Render.indent: empty stack")

    def render(self, doc: Document, error_ok: bool = False):
        """Render a document on the stream.

        Parameters
        ----------
        doc : Document
            Document to render

        error_ok : bool, optional
            If True, do not raise an exception if an error occurs and returns. Defaults to False.

        Raises
        ------
        ScadeOneException
            Raised if there is no stream to render on, or an error is raised
        """
        if not self._stream:
            raise ScadeOneException("No stream to render on")
        if not isinstance(doc, Document):
            raise ScadeOneException("Renderer: Document expected")
        try:
            self._render(doc.start_doc)
        except Exception as e:
            if not error_ok:
                raise ScadeOneException(f"Error during rendering: ({e})") from e

    def _write(self, obj: Any):
        """Write obj on stream. str(obj) shall be a string without newlines"""
        text = str(obj)
        if text.find("\n") != -1:
            raise ScadeOneException(f"Renderer: text '{text}' shall not contain newlines")
        self._col += len(text)
        self._stream.write(text)  # type: ignore # (stream is IO)

    def _nl(self, with_indent: bool):
        self._stream.write("\n")  # type: ignore # (stream is IO)
        self._line += 1
        self._col = 0
        if with_indent:
            self._write(" " * self.curr_indent)

    def _push_indent(self):
        indent = self.curr_indent + self.default_indent
        self._indent_stack.append(indent)

    def _pop_indent(self):
        try:
            self._indent_stack.pop()
        except Exception:  # pylint: disable=broad-except
            self._indent_stack = [self._initial_indent]

    def _push_mark(self):
        self._indent_stack.append(self._col)

    def _render(self, doc: DElt):
        def _get_render_fn(doc) -> Callable[["Renderer", DElt], None]:
            "Auxiliary function to get the render function, which is a method of the Renderer class"
            class_name = doc.__class__.__name__
            func = getattr(self, f"_render_{class_name}", self._render_NoFunc)
            return func

        _current_doc = doc
        while _current_doc:
            func = _get_render_fn(_current_doc)
            func(_current_doc)
            _current_doc = _current_doc.next

    def _render_NoFunc(self, doc):  # pylint: disable=C0103
        class_name = doc.__class__.__name__
        raise ScadeOneException(f"Render._render_{class_name}() does not exist")

    def _render_DElt(self, _: DElt):  # pylint: disable=C0103
        # Default render: do nothing
        pass

    def _render_DText(self, doc: DText):  # pylint: disable=C0103
        self._write(doc.string)

    def _render_DBlock(self, doc: DBlock):  # pylint: disable=C0103
        self._render(doc.doc)

    def _render_DLineBreak(self, doc: DLineBreak):  # pylint: disable=C0103
        self._nl(doc.with_indent)

    def _render_DIndent(self, doc: DIndent):  # pylint: disable=C0103
        if doc.indent is EIndentation.INDENT:
            self._push_indent()
        elif doc.indent is EIndentation.UNINDENT:
            self._pop_indent()
        elif doc.indent is EIndentation.MARK:
            self._push_mark()


if __name__ == "__main__":
    import doctest

    doctest.testmod()
